@@ -35,11 +35,19 @@ void
 taint_env()
 {
     SV** svp;
-    MAGIC *mg;
+    MAGIC* mg;
+    char** e;
+    static char* misc_env[] = {
+	"IFS",		/* most shells' inter-field separators */
+	"ENV",		/* ksh dain bramage #1 */
+	"CDPATH",	/* ksh dain bramage #2 */
+	"TERM",		/* some termcap libraries' dain bramage */
+	NULL
+    };
 
 #ifdef VMS
     int i = 0;
-    char name[14] = "DCL$PATH";
+    char name[10 + TYPE_DIGITS(int)] = "DCL$PATH";
 
     while (1) {
 	if (i)
@@ -71,9 +79,11 @@ taint_env()
 	}
     }
 
-    svp = hv_fetch(GvHVn(envgv),"IFS",3,FALSE);
-    if (svp && *svp != &sv_undef && SvTAINTED(*svp)) {
-	TAINT;
-	taint_proper("Insecure %s%s", "$ENV{IFS}");
+    for (e = misc_env; *e; e++) {
+	svp = hv_fetch(GvHVn(envgv), *e, strlen(*e), FALSE);
+	if (svp && *svp != &sv_undef && SvTAINTED(*svp)) {
+	    TAINT;
+	    taint_proper("Insecure $ENV{%s}%s", *e);
+	}
     }
 }
