@@ -4,6 +4,7 @@
 typedef struct {
 int	signature_begin;
 int *	(*pfnerrno)(void);
+char ***(*pfnenviron)(void);
 FILE*	(*pfnstdin)(void);
 FILE*	(*pfnstdout)(void);
 FILE*	(*pfnstderr)(void);
@@ -51,6 +52,9 @@ int	(*pfnopenmode)(int mode);
 int	(*pfn_open_osfhandle)(long handle, int flags);
 long	(*pfn_get_osfhandle)(int fd);
 int	(*pfnspawnvpe)(int mode, const char *cmdname, const char *const *argv, const char *const *envp);
+int	(*pfnmkdir)(const char *path);
+int	(*pfnrmdir)(const char *path);
+int	(*pfnchdir)(const char *path);
 int		signature_end;
 } WIN32_IOSUBSYSTEM; 
 
